from ratio.router.router import Router
from ratio.http.request import Request
from ratio.application.discover_routes import discover_routes
from asgiref.typing import (
    Scope,
    ASGIReceiveCallable,
    ASGISendCallable,
    HTTPResponseBodyEvent,
    HTTPResponseStartEvent,
)
import pathlib


class Ratio:
    application_root: pathlib.Path
    router: Router

    def __init__(self, root: pathlib.Path | None):
        self.application_root = root if root is not None else pathlib.Path.cwd()

        routes = discover_routes(self.application_root)
        self.router = Router(routes)

    async def __call__(
        self, scope: Scope, receive: ASGIReceiveCallable, send: ASGISendCallable
    ) -> None:
        if scope["type"] != "http":
            raise NotImplementedError

        # Temporarily use the receive variable here, to satisfy Ruff
        # TODO: remove this when we actually perform more ASGI logic, incorporating receive.
        _receive = receive

        request = Request.from_http_scope(scope)
        route = await self.router.resolve(request)
        response = await route()

        start: HTTPResponseStartEvent = {
            "type": "http.response.start",
            "status": response.code,
            "headers": [
                (b"content-type", b"text/plain"),
            ],
            "trailers": False,
        }

        await send(start)

        body: HTTPResponseBodyEvent = {
            "type": "http.response.body",
            "body": str.encode(response.message),
            "more_body": False,
        }

        await send(body)



import pathlib
import importlib.util
from types import ModuleType


def load_local_module(name: str, path: pathlib.Path) -> ModuleType | None:
    spec = importlib.util.spec_from_file_location(name, path)

    if spec is None or spec.loader is None:
        return None

    module = importlib.util.module_from_spec(spec)

    if module is None:
        return None

    spec.loader.exec_module(module)

    return module



from typing import TypeVar, TypeGuard, Type
from ratio.router.route import Route

_T = TypeVar("_T", bound=type)


def is_route(value: _T) -> TypeGuard[Type[Route]]:
    try:
        return issubclass(value, Route) and not value == Route
    except TypeError:
        return False



from ratio.application.load_local_module import load_local_module
from ratio.router.route import Route
from ratio.router.route_path import RoutePath
from ratio.router.is_route import is_route
import pathlib
from typing import Type


def map_discovered_route_to_route(
    path: pathlib.Path,
) -> Type[Route] | None:
    resolved_path = path.resolve()

    module = load_local_module("route", resolved_path)
    module_attributes = [getattr(module, item) for item in dir(module)]
    potential_routes = [
        attribute for attribute in module_attributes if is_route(attribute)
    ]

    if not len(potential_routes) == 1:
        return None

    return potential_routes[0]


def discover_routes(
    application_root: pathlib.Path,
) -> dict[RoutePath, Type[Route]]:
    routes_directory = application_root / "routes"
    if not routes_directory.exists() or not routes_directory.is_dir():
        return {}

    potential_routes = {
        RoutePath.from_path(
            path.relative_to(routes_directory).with_suffix("")
        ): map_discovered_route_to_route(path)
        for path in routes_directory.glob("**/*.py")
    }

    return {
        route_path: path
        for route_path, path in potential_routes.items()
        if path is not None
    }
